@@ -2,6 +2,7 @@
 #include "util.hpp"
 #include <md5.h>
 #include <random>
+#include <sys/fsuid.h>
 static const constexpr auto key_or_plain = [](auto && lhs, auto && rhs) {
 	static const constexpr auto key =
 	    vore::overload{[](const std::string_view & s) { return s; }, [](const std::pair<std::string_view, std::string_view> & kv) { return kv.first; }};
@@ -161,12 +162,15 @@ struct Job {
 	std::string jobid;
 	std::string unit_name;
 	std::string_view user;
-	std::optional<std::string_view> home;  // 'static
+	std::optional<std::string_view> user_home;  // 'static
+	uid_t user_uid;
+	gid_t user_gid;
 	cron_mail_success_t cron_mail_success;
 	cron_mail_format_t cron_mail_format;
 	struct {
-		vore::span<const std::string_view *> command;  // subview of parts
-		std::optional<std::string> command0;           // except this is command[0] if set
+		vore::span<std::string_view *> command;  // subview of parts
+		std::optional<std::string> command0;     // except this is command[0] if set
+		bool nopercent;
 
 		struct command_iter {
 			using iterator_category = std::input_iterator_tag;
@@ -175,7 +179,7 @@ struct Job {
 			using pointer           = const std::string_view *;
 			using reference         = const std::string_view &;
 
-			vore::span<const std::string_view *> command;
+			vore::span<std::string_view *> command;
 			std::optional<std::string_view> command0;
 
 			command_iter & operator++() noexcept {
@@ -209,6 +213,7 @@ struct Job {
 		}
 	} command;
 	std::string execstart;
+	std::string_view input_data;
 	bool valid;
 
 	Job(std::string_view filename, std::string_view line) {
@@ -225,7 +230,10 @@ struct Job {
 		this->random_delay    = 0;
 		this->persistent      = false;
 		this->user            = "root"sv;
+		this->user_uid        = 0;
+		this->user_gid        = 0;
 		this->command         = {};
+		this->input_data      = {};
 		this->valid           = true;
 		this->batch           = false;
 		this->sunday_is_seven = false;
@@ -339,6 +347,7 @@ struct Job {
 			this->log(Log::WARNING, "invalid DELAY");
 		this->command.command.b = &*(this->parts.begin() + 3);
 		this->command.command.e = &*(this->parts.end());
+		this->command.nopercent = true;
 	}
 
 	// crontab --translate <something>
@@ -391,6 +400,7 @@ struct Job {
 		}
 		this->command.command.e = &*(this->parts.end());
 		this->jobid             = (std::string{this->basename} += '-') += this->user;
+		this->parse_percent();
 	}
 
 	// 6 2 * * * (user) do something
@@ -417,9 +427,9 @@ struct Job {
 			days = "7/1"sv;
 			dows = dows.substr(0, 1);
 		}
-		this->timespec_dom    = this->parse_time_unit<false, std::uint8_t>(days, "day", DAYS_SET, DAYS_RANGE, int_map, this->timespec_dom_raw);
-		this->timespec_dow    = this->parse_time_unit<true, std::string_view>(dows, "dow", DOWS_SET, DOWS_RANGE, dow_map, this->timespec_dow_raw);
-		this->timespec_month  = this->parse_time_unit<false, std::uint8_t>(months, "month", MONTHS_SET, MONTHS_RANGE, month_map, this->timespec_month_raw);
+		this->timespec_dom   = this->parse_time_unit<false, std::uint8_t>(days, "day", DAYS_SET, DAYS_RANGE, int_map, this->timespec_dom_raw);
+		this->timespec_dow   = this->parse_time_unit<true, std::string_view>(dows, "dow", DOWS_SET, DOWS_RANGE, dow_map, this->timespec_dow_raw);
+		this->timespec_month = this->parse_time_unit<false, std::uint8_t>(months, "month", MONTHS_SET, MONTHS_RANGE, month_map, this->timespec_month_raw);
 
 		this->sunday_is_seven = dows.back() == '7' || [&] {
 			if(dows.size() < 3)
@@ -443,6 +453,24 @@ struct Job {
 		}
 		this->command.command.e = &*(this->parts.end());
 		this->jobid             = (std::string{this->basename} += '-') += this->user;
+		this->parse_percent();
+	}
+
+	// For non-anacron jobs: find the first unescaped %, terminate this->command there, and put the rest into this->input_data
+	auto parse_percent() -> void {
+		for(auto itr = this->command.command.b; itr != this->command.command.e; ++itr) {
+			std::size_t cur{};
+			for(std::size_t percent; (percent = itr->find('%', cur)) != std::string_view::npos;) {
+				if(percent == 0 || (*itr)[percent - 1] != '\\') {
+					this->input_data = {itr->data() + percent + 1, this->line.data() + this->line.size()};
+					itr->remove_suffix(itr->size() - percent);
+					this->command.command.e = itr + 1;
+					return;
+				}
+
+				cur = percent + 1;
+			}
+		}
 	}
 
 	static const constexpr std::uint8_t TIMESPEC_ASTERISK = -1;
@@ -496,21 +524,9 @@ struct Job {
 			return;
 
 
-		if(!this->home) {
-			static std::map<std::string, std::string, std::less<>> pwnam_cache;
-			auto itr = pwnam_cache.find(this->user);
-			if(itr == std::end(pwnam_cache)) {
-				std::string user{this->user};
-				if(auto ent = getpwnam(user.data()))
-					itr = pwnam_cache.emplace(std::move(user), ent->pw_dir).first;
-			}
-			if(itr != std::end(pwnam_cache))
-				this->home = itr->second;
-		}
-
-		if(this->home) {
+		if(this->pull_user()) {
 			if(this->command[0].starts_with("~/"sv)) {
-				this->command.command0 = std::string{ * this->home} += this->command[0].substr(1);
+				this->command.command0 = std::string{ * this->user_home} += this->command[0].substr(1);
 				++this->command.command.b;
 			}
 
@@ -518,7 +534,7 @@ struct Job {
 				if(itr->second.starts_with("~/") || itr->second.find(":~/"sv) != std::string_view::npos) {
 					for(auto && path : vore::soft_tokenise{itr->second, ":"sv}) {
 						if(path.starts_with("~/")) {
-							this->environment_PATH_storage += *this->home;
+							this->environment_PATH_storage += *this->user_home;
 							this->environment_PATH_storage += path.substr(1);
 						} else
 							this->environment_PATH_storage += path;
@@ -530,6 +546,26 @@ struct Job {
 		}
 	}
 
+	auto pull_user() -> bool {
+		if(this->user_home)
+			return true;
+
+		static std::map<std::string, std::tuple<std::string, uid_t, gid_t>, std::less<>> pwnam_cache;
+		auto itr = pwnam_cache.find(this->user);
+		if(itr == std::end(pwnam_cache)) {
+			std::string user{this->user};
+			if(auto ent = getpwnam(user.data()))
+				itr = pwnam_cache.emplace(std::move(user), std::tuple{ent->pw_dir, ent->pw_uid, ent->pw_gid}).first;
+		}
+		if(itr != std::end(pwnam_cache)) {
+			this->user_home = std::get<0>(itr->second);
+			this->user_uid  = std::get<1>(itr->second);
+			this->user_gid  = std::get<2>(itr->second);
+		}
+
+		return !!this->user_home;
+	}
+
 	auto is_active() -> bool {
 		if(this->schedule == "reboot"sv && !access(REBOOT_FILE, F_OK))
 			return false;
@@ -693,7 +729,7 @@ struct Job {
 			ADDBOTH('-');
 		}
 		timespec_comma(this->timespec_dom, this->timespec_dom_raw);
-		if (this->last_dow) {
+		if(this->last_dow) {
 			ADDBOTH('/');
 			ADDBOTH('1');
 		}
@@ -704,15 +740,47 @@ struct Job {
 		ADDBOTH(":00"sv);
 	}
 
+	template <class F>
+	auto debackslashpercentise(std::string_view cmd, F && append) -> void {
+		if(this->command.nopercent) {
+			append(cmd, false);
+			return;
+		}
+
+		while(cmd.size()) {
+			auto backpct = cmd.find("\\%"sv);
+			if(backpct == std::string_view::npos) {
+				append(cmd, true);
+				cmd = {};
+			} else {
+				append(cmd.substr(0, backpct), true);
+				append("%"sv, false);
+				cmd.remove_prefix(backpct + 2);
+			}
+		}
+	}
+
 	auto generate_scriptlet() -> std::optional<std::string> {
 		// ...only if needed
 		assert(!this->unit_name.empty());
-		if(this->command.size() == 1) {
+		if(this->command.size() == 1 && this->pull_user()) {
+			this->debackslashpercentise(this->command[0], [&](auto && segment, auto) { this->execstart += segment; });
+
+			uid_t uid;
+			gid_t gid;
+			if(this->user_uid)
+				uid = setfsuid(this->user_uid);
+			if(this->user_gid)
+				gid = setfsgid(this->user_gid);
 			struct stat sb;
-			if(!stat(MAYBE_DUPA(this->command[0]), &sb) && S_ISREG(sb.st_mode)) {
-				this->execstart = this->command[0];
+			auto statres = stat(this->execstart.c_str(), &sb);
+			if(this->user_uid)
+				setfsuid(uid);
+			if(this->user_gid)
+				setfsgid(gid);
+
+			if(!statres && S_ISREG(sb.st_mode))
 				return {};
-			}
 		}
 
 		auto scriptlet  = ((std::string{TARGET_DIR} += '/') += this->unit_name) += ".sh"sv;
@@ -773,9 +841,9 @@ struct Job {
 	auto generate_service(FILE * into) -> void {
 		this->generate_unit_header(into, "Cron");
 		if(auto itr = this->environment.find("MAILTO"sv); itr != std::end(this->environment) && itr->second.empty())
-			;  // mails explicitely disabled
+			;  // mails explicitly disabled
 		else if(!HAS_SENDMAIL)
-			;  // mails automaticaly disabled
+			;  // mails automatically disabled
 		else {
 			this->format_on_failure(into, "Failure");
 
@@ -792,8 +860,8 @@ struct Job {
 		}
 		if(this->user != "root"sv || this->filename.find(STATEDIR) != std::string_view::npos) {
 			std::fputs("Requires=systemd-user-sessions.service\n", into);
-			if(this->home)
-				std::fprintf(into, "RequiresMountsFor=%.*s\n", FORMAT_SV(*this->home));
+			if(this->user_home)
+				std::fprintf(into, "RequiresMountsFor=%.*s\n", FORMAT_SV(*this->user_home));
 		}
 		std::fputc('\n', into);
 
@@ -814,6 +882,23 @@ struct Job {
 			environment_write(this->environment, into);
 			std::fputc('\n', into);
 		}
+		if(!this->input_data.empty()) {
+			std::fputs("StandardInput=data\n", into);
+			std::fputs("StandardInputData=", into);
+			b64 data{into};
+
+			this->debackslashpercentise(this->input_data, [&](auto segment, auto replace_percent) {
+				if(replace_percent)
+					for(std::size_t pct; (pct = segment.find('%')) != std::string_view::npos;) {
+						data.feed(segment.substr(0, pct));
+						data.feed("\n"sv);
+						segment.remove_prefix(pct + 1);
+					}
+				data.feed(segment);
+			});
+			data.finish();
+			std::fputc('\n', into);
+		}
 		if(this->batch) {
 			std::fputs("CPUSchedulingPolicy=idle\n", into);
 			std::fputs("IOSchedulingClass=idle\n", into);
@@ -883,7 +968,7 @@ struct Job {
 					continue;
 				if(!std::exchange(first, false))
 					std::fputc(' ', f);
-				std::fwrite(hunk.data(), 1, hunk.size(), f);
+				this->debackslashpercentise(hunk, [&](auto && segment, auto) { std::fwrite(segment.data(), 1, segment.size(), f); });
 			}
 			if(!first)
 				std::fputc('\n', f);
@@ -904,7 +989,8 @@ struct Job {
 		if(symlink(timer.c_str(), (((std::string{TIMERS_DIR} += '/') += this->unit_name) += ".timer"sv).c_str()) == -1 && errno != EEXIST)
 			return output_err(timer, "link"), false;
 
-		auto service = ((std::string{TARGET_DIR} += '/') += this->unit_name) += ".service"sv;
+		auto & service = timer;
+		service.replace(service.size() - "timer"sv.size(), service.size(), "service"sv);
 		{
 			vore::file::FILE<false> s{service.c_str(), "we"};
 			if(!s)
@@ -946,9 +1032,9 @@ static auto environment_write(const std::map<std::string_view, std::string_view>
 }
 
 
-template <class F>
+template <class F, class G = void (*)(Job &&)>
 static auto parse_crontab(std::string_view filename, withuser_t withuser, bool anacrontab, cron_mail_success_t default_cron_mail_success,
-                          cron_mail_format_t default_cron_mail_format, F && cbk) -> bool {
+                          cron_mail_format_t default_cron_mail_format, F && cbk, std::optional<G> && preview_env = std::nullopt) -> bool {
 	vore::file::mapping map;
 	{
 		vore::file::fd<true> f{filename.data(), O_RDONLY | O_CLOEXEC};
@@ -1010,6 +1096,11 @@ static auto parse_crontab(std::string_view filename, withuser_t withuser, bool a
 		j.generate_schedule();
 		cbk(j);
 	}
+	if(preview_env) {
+		Job j{filename, ""sv};
+		j.decode_environment(environment, /*default_persistent=*/false, default_cron_mail_success, default_cron_mail_format);
+		(*preview_env)(std::move(j));
+	}
 	return true;
 }
 
@@ -1183,7 +1274,8 @@ static auto is_backup(const char * path, const std::string_view & name) -> bool
 	if(name == ".placeholder"sv)
 		return true;
 
-	bool backup = name[0] == '.' || name.find('~') != std::string_view::npos || name.find(".dpkg-"sv) != std::string_view::npos || name.find(".rpm"sv) != std::string_view::npos || name == "0anacron"sv;
+	bool backup = name[0] == '.' || name.find('~') != std::string_view::npos || name.find(".dpkg-"sv) != std::string_view::npos ||
+	              name.find(".rpm"sv) != std::string_view::npos || name == "0anacron"sv;
 	if(backup)
 		log(Log::DEBUG, "ignoring %s/%.*s", path, FORMAT_SV(name));
 	return backup;
@@ -1197,34 +1289,38 @@ static auto realmain() -> int {
 
 	std::optional<std::string> fallback_mailto;
 	std::map<std::string_view, std::uint8_t> distro_start_hour;
-	cron_mail_success_t toplevel_cron_mail_success = cron_mail_success_t::dflt;
-	cron_mail_format_t toplevel_cron_mail_format   = cron_mail_format_t::dflt;
+	auto toplevel_cron_mail_success = cron_mail_success_t::dflt;
+	auto toplevel_cron_mail_format  = cron_mail_format_t::dflt;
 
-	if(!parse_crontab("/etc/crontab", withuser_t::from_cmd0, /*anacrontab=*/false, cron_mail_success_t::dflt, cron_mail_format_t::dflt, [&](auto && job) {
-		   if(auto itr = job.environment.find("MAILTO"sv); itr != std::end(job.environment))
-			   fallback_mailto = itr->second;
-		   if(!job.valid) {
-			   log(Log::ERR, "truncated line in /etc/crontab: %.*s", FORMAT_SV(job.line));
-			   return;
-		   }
+	if(!parse_crontab(
+	       "/etc/crontab", withuser_t::from_cmd0, /*anacrontab=*/false, cron_mail_success_t::dflt, cron_mail_format_t::dflt,
+	       [&](auto && job) {
+		       if(!job.valid) {
+			       log(Log::ERR, "truncated line in /etc/crontab: %.*s", FORMAT_SV(job.line));
+			       return;
+		       }
 
-		   toplevel_cron_mail_success = job.cron_mail_success;
-		   toplevel_cron_mail_format  = job.cron_mail_format;
+		       // legacy boilerplate: ignore jobs that run /etc/cron.hourly, daily, weekly, monthly
+		       //                     (but save the starting hour for daily, weekly, and monthly)
+		       if(job.line.find("/etc/cron.hourly"sv) != std::string_view::npos)
+			       return;
+		       for(auto && disableable_period : {"daily"sv, "weekly"sv, "monthly"sv})
+			       if(job.line.find("/etc/cron."s += disableable_period) != std::string_view::npos) {
+				       if(auto hour = *job.timespec_hour.begin(); hour != Job::TIMESPEC_ASTERISK)
+					       distro_start_hour[disableable_period] = hour;
 
-		   // legacy boilerplate: ignore jobs that run /etc/cron.hourly, daily, weekly, monthly
-		   //                     (but save the starting hour for daily, weekly, and monthly)
-		   if(job.line.find("/etc/cron.hourly"sv) != std::string_view::npos)
-			   return;
-		   for(auto && disableable_period : {"daily"sv, "weekly"sv, "monthly"sv})
-			   if(job.line.find("/etc/cron."s += disableable_period) != std::string_view::npos) {
-				   if(auto hour = *job.timespec_hour.begin(); hour != Job::TIMESPEC_ASTERISK)
-					   distro_start_hour[disableable_period] = hour;
+				       return;
+			       }
 
-				   return;
-			   }
+		       generate_timer_unit(job);
+	       },
+	       std::make_optional([&](auto && envjob) {
+		       if(auto itr = envjob.environment.find("MAILTO"sv); itr != std::end(envjob.environment))
+			       fallback_mailto = itr->second;
 
-		   generate_timer_unit(job);
-	   }))
+		       toplevel_cron_mail_success = envjob.cron_mail_success;
+		       toplevel_cron_mail_format  = envjob.cron_mail_format;
+	       })))
 		log(Log::ERR, "%s: %s", "/etc/crontab", std::strerror(errno));
 
 
@@ -1265,7 +1361,7 @@ static auto realmain() -> int {
 				job.period            = period;
 				job.start_hour        = distro_start_hour[period];  // default 0
 				job.boot_delay        = i * 5;
-				job.command           = {{&command, &command + 1}, {}};
+				job.command           = {{&command, &command + 1}, {}, true};
 				job.jobid             = (std::string{period} += '-') += basename;
 				job.cron_mail_success = toplevel_cron_mail_success;
 				job.cron_mail_format  = toplevel_cron_mail_format;
@@ -1289,7 +1385,7 @@ static auto realmain() -> int {
 		log(Log::ERR, "%s: %s", "/etc/anacrontab", std::strerror(errno));
 
 	if(struct stat sb; !stat(STATEDIR, &sb) && S_ISDIR(sb.st_mode)) {
-		// /var is avaible
+		// /var is available
 		for_each_file(STATEDIR, [&](std::string_view basename) {
 			if(basename.find('.') != std::string_view::npos)
 				return;
